package checkers.igj;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.lang.model.element.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;

import checkers.igj.quals.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.*;

/**
 * Generates {@link AnnotatedTypeMirror} instances accounting for the
 * properties of the IGJ type system.  It add implicit and default IGJ
 * annotations, only if the user does not annotate the type explicitly.
 *
 * This factory will add the {@link Immutable} annotation to a type if the
 * input is
 * <ul>
 *  <li>a primitive type,</li>
 *  <li>a known immutable type, if the class type is annotated as
 *      {@code Immutable}, or</li>
 *  <li>an enumerator type.</li>
 * </ul>
 *
 * It will add the {@link ReadOnly} annotation to a type if the input is
 * <ul>
 *  <li>a method receiver for an immutable class</li>
 *  <li>a result of unification of different immutabilities (e.g.
 *      within Conditional Expressions)</li>
 * </ul>
 *
 * It will add the {@link Mutable} annotation to a type if the input is
 * a use of a mutable type.
 *
 * As an implementation detail, it adds a special annotation, IGJPlaceHolder,
 *
 * Furthermore, it resolves {@link I} annotation to the proper annotation
 * according to its specification.
 */
public class IGJAnnotatedTypeFactory extends AnnotatedTypeFactory{
    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The various IGJ annotations. */
    private final AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I,
            PLACE_HOLDER, ASSIGNS_FIELDS;

    private final AnnotationUtils annoUtils;

    /**
     * Constructor for IGJAnnotatedTypeFactory object.
     *
     * @param checker the checker to which this factory belongs
     * @param root  the compilation unit the annotation processor is
     *              processing currently
     */
    public IGJAnnotatedTypeFactory(IGJChecker checker,
            CompilationUnitTree root) {
        super(checker, root);
        treePre = new TreePreAnnotator();
        typePost = new TypePostAnnotator();
        AnnotationFactory annoFactory = new AnnotationFactory(env);
        READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        IMMUTABLE = annoFactory.fromName(Immutable.class.getCanonicalName());
        I = annoFactory.fromName(I.class.getCanonicalName());
        PLACE_HOLDER =
            annoFactory.fromName(IGJPlaceHolder.class.getCanonicalName());
        ASSIGNS_FIELDS =
            annoFactory.fromName(AssignsFields.class.getCanonicalName());
        annoUtils = new AnnotationUtils(env);
    }

    /**
     * Returns the annotation specifying the immutability type of {@code type}.
     */
    private AnnotationMirror getImmutabilityAnnotation(AnnotatedTypeMirror type) {
        if (type.hasAnnotation(I))
            return type.getAnnotation(I.class.getCanonicalName());
        return super.annoRelations.getAnnotation(type.getAnnotations());
    }

    /**
     * @param type  an annotated type mirror
     * @return  true iff the type is specified an immutability type,
     *          false otherwise
     */
    private boolean hasImmutabilityAnnotation(AnnotatedTypeMirror type) {
        return getImmutabilityAnnotation(type) != null;
    }

    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treePre.visit(tree, type);
        typePost.visit(type, tree.getKind() == Tree.Kind.CLASS);
    }

    @Override
    protected void annotateImplicit(Element element, AnnotatedTypeMirror type) {
        // Add annotations based on type properties.
        typePost.visit(type, element.getKind().isClass()
                || element.getKind().isInterface());
    }

    /**
     * Infers the immutability of {@code @I}s based on the provided types, and
     * replace all instances of {@code @I} with their corresponding qualifiers.
     *
     * @param type      the type with {@code @I} annotation
     * @param provided  the types with qualifiers bound to {@code @I}
     * @return true iff a qualifier has been resolved.
     */
    private boolean resolveImmutabilityTypeVar(AnnotatedTypeMirror type,
            AnnotatedTypeMirror ...provided) {
        ImmutabilityTemplateCollector collector =
            new ImmutabilityTemplateCollector();

        Map<String, AnnotationMirror> templateMapping = null;

        for (AnnotatedTypeMirror pt : provided)
            templateMapping = collector.reduce(templateMapping, collector.visit(pt));

        // There is nothing to resolve
        if (templateMapping == null || templateMapping.isEmpty())
            return false;

        new ImmutabilityResolver().visit(type, templateMapping);
        return true;
    }

    /**
     * Replace all instances of {@code @I} in the super types with the
     * immutability of the current type
     *
     * @param type  the type whose supertypes are requested
     * @param supertypes    the supertypes of type
     */
    @Override
    protected void postDirectSuperTypes(AnnotatedTypeMirror type,
            List<? extends AnnotatedTypeMirror> supertypes) {
        Map<String, AnnotationMirror> templateMapping =
            new ImmutabilityTemplateCollector().visit(type);

        new ImmutabilityResolver().visit(supertypes, templateMapping);
        typePost.visit(supertypes, false);
    }

    /**
     * Resolve the instances of {@code @I} in the {@code elementType} based
     * on {@code owner}, according to is specification.
     *
     */
    @Override
    public void postAsMemberOf(AnnotatedTypeMirror elementType,
            AnnotatedTypeMirror owner, Element element) {
        resolveImmutabilityTypeVar(elementType, owner);
    }

    /**
     * Override the default behavior of implicitly adding annotations on the
     * class type to the declared types within the passed type.
     */
    @Override
    protected void annotateInheritedFromClass(AnnotatedTypeMirror type) {
        // Do nothing
    }

    /**
     * Returns the type and field {@code this},  for the scope of this tree.
     * In IGJ, the self type is the method receiver in this scope.
     */
    @Override
    public AnnotatedDeclaredType getSelfType(Tree tree) {
        AnnotatedDeclaredType act = getCurrentClassType(tree);
        AnnotatedDeclaredType methodReceiver = getCurrentMethodReceiver(tree);

        if (methodReceiver == null)
            return act;
        // Are we in a mutable or Immutable scope
        if (methodReceiver.hasAnnotation(MUTABLE) ||
                methodReceiver.hasAnnotation(IMMUTABLE)) {
            return methodReceiver;
        } else if (act.hasAnnotation(I) || act.hasAnnotation(IMMUTABLE)) {
            if (methodReceiver.hasAnnotation(ASSIGNS_FIELDS))
                act.addAnnotation(ASSIGNS_FIELDS);
            return act;
        } else // TODO: Figure out when
            return methodReceiver;
    }

    @Override
    public Collection<AnnotationMirror> unify(Collection<AnnotationMirror> c1,
            Collection<AnnotationMirror> c2) {
        assert c1.size() == 1 && c2.size() == 1;
        AnnotationMirror anno1 = c1.iterator().next();
        AnnotationMirror anno2 = c2.iterator().next();

        if (AnnotationUtils.isSame(anno1, anno2))
            return Collections.singleton(anno1);
        else if (AnnotationUtils.isSame(anno1, PLACE_HOLDER))
            return c2;
        else if (AnnotationUtils.isSame(anno2, PLACE_HOLDER))
            return c1;
        else if (annoRelations.isSubtype(anno1, anno2))
            return Collections.singleton(anno1);
        else if (annoRelations.isSubtype(anno2, anno1))
            return Collections.singleton(anno2);
        else
            return Collections.singleton(READONLY);
    }

    /**
     * A helper class that resolves the immutability on a types based on a
     * provided mapping.
     *
     * It returns a set of the annotations that were inserted. This is important
     * to recognize which immutability type variables were resolved and which
     * are to be made into place holder.
     */
    private class ImmutabilityResolver extends
    AnnotatedTypeScanner<Void, Map<String, AnnotationMirror>> {

        public void visit(Iterable<? extends AnnotatedTypeMirror> types,
                Map<String, AnnotationMirror> templateMapping) {
            if (templateMapping != null && !templateMapping.isEmpty()) {
                for (AnnotatedTypeMirror type : types)
                    visit(type, templateMapping);
            }
        }

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type,
                Map<String, AnnotationMirror> p) {
            if (type.hasAnnotation(I)) {
                String immutableString =
                    annoUtils.parseStringValue(getImmutabilityAnnotation(type),
                            IGJChecker.IMMUTABILITY_KEY);
                if (p.containsKey(immutableString)) {
                    type.removeAnnotation(I);
                    type.addAnnotation(p.get(immutableString));
                }
            }

            return super.visitDeclared(type, p);
        }
    }

    /**
     * Helper class to annotate trees.
     *
     * Currently it only adds an PLACE_HOLDER for new classes and new arrays,
     * when an annotation is not specified
     */
    private class TreePreAnnotator
    extends SimpleTreeVisitor<Void, AnnotatedTypeMirror> {

        @Override
        public Void visitNewClass(NewClassTree node, AnnotatedTypeMirror p) {
            if (!hasImmutabilityAnnotation(p)) {
                // If no immutability is specified, use a place holder
                AnnotatedTypeMirror ct = fromElement(
                        ((AnnotatedDeclaredType)p).getUnderlyingType().asElement());

                if (!hasImmutabilityAnnotation(ct) || ct.hasAnnotation(I))
                    p.addAnnotation(PLACE_HOLDER);
            }

            return null;
        }

        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror p) {
            if (!node.getName().contentEquals("this"))
                resolveImmutabilityTypeVar(p, getReceiver(node));

            return null;
        }
    }

    /**
     * Helper class for annotating unannotated types.
     */
    private class TypePostAnnotator extends AnnotatedTypeScanner<Void, Boolean> {
        public Void visit(Iterable<? extends AnnotatedTypeMirror> types, Boolean p) {
            for (AnnotatedTypeMirror type : types)
                visit(type, p);
            return null;
        }
        /** All primitives are immutable **/
        @Override
        public Void visitPrimitive(AnnotatedPrimitiveType type, Boolean p) {
            if (!hasImmutabilityAnnotation(type))
                type.addAnnotation(PLACE_HOLDER);
            return super.visitPrimitive(type, p);
        }

        /**
         * For Declared types:
         *  Classes are mutable
         *  Interface declaration are placeholders
         *  Enum and annotations  are immutable
         */
        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Boolean p) {
            if (!hasImmutabilityAnnotation(type)) {
                // Actual element
                TypeElement element = (TypeElement)type.getUnderlyingType().asElement();
                AnnotatedDeclaredType elementType = fromElement(element);

                if (TypesUtils.isBoxedPrimitive(type.getUnderlyingType()))
                    type.addAnnotation(PLACE_HOLDER);
                else if (ElementUtils.isObject(element))
                    type.addAnnotation(PLACE_HOLDER);
                else if (elementType.hasAnnotation(IMMUTABLE))
                    type.addAnnotation(IMMUTABLE);
                else if (hasImmutabilityAnnotation(elementType)) // not immutable
                    type.addAnnotation(MUTABLE);
                else if (element.getKind() == ElementKind.CLASS)
                    type.addAnnotation(MUTABLE);
                else if (element.getKind() == ElementKind.INTERFACE)
                    type.addAnnotation(p ? PLACE_HOLDER : MUTABLE);
                else if (element.getKind() == ElementKind.ENUM ||
                        element.getKind() == ElementKind.ANNOTATION_TYPE)
                    type.addAnnotation(IMMUTABLE);
                else
                    assert false : "shouldn't be here!";
            }
            return super.visitDeclared(type, p);
        }

        /** null are subtypes of everything **/
        @Override
        public Void visitNull(AnnotatedNullType type, Boolean p) {
            type.addAnnotation(PLACE_HOLDER);
            return super.visitNull(type, p);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Boolean p) {
            if (hasImmutabilityAnnotation(type.getReceiverType()))
                return super.visitExecutable(type, p);

            AnnotatedDeclaredType receiver = type.getReceiverType();
            TypeElement ownerElement = ElementUtils.enclosingClass(type.getElement());
            AnnotatedDeclaredType ownerType = getAnnotatedType(ownerElement);

            if (ElementUtils.isObject(ownerElement)) {
                receiver.addAnnotation(PLACE_HOLDER);
            } else if (type.getElement().getKind() == ElementKind.CONSTRUCTOR) {
                receiver.addAnnotation(
                        ownerType.hasAnnotation(MUTABLE) ? MUTABLE : ASSIGNS_FIELDS);
            } else if (ownerType.hasAnnotation(IMMUTABLE)) {
                receiver.addAnnotation(READONLY);
            } else {
                receiver.addAnnotation(MUTABLE);
            }

            return super.visitExecutable(type, p);
        }

        @Override
        public Void visitArray(AnnotatedArrayType type, Boolean p) {
            if (!hasImmutabilityAnnotation(type))
                type.addAnnotation(MUTABLE);
            return super.visitArray(type, p);
        }
    }

    /**
     * A Helper class that tries to resolve the immutability type variable,
     * as the type variable is assigned to the most restricted immutability
     */
    private class ImmutabilityTemplateCollector
    extends SimpleAnnotatedTypeVisitor<Map<String, AnnotationMirror>, AnnotatedTypeMirror> {

        public Map<String, AnnotationMirror> reduce(Map<String, AnnotationMirror> r1,
                Map<String, AnnotationMirror> r2) {
            Map<String, AnnotationMirror> result =
                new HashMap<String, AnnotationMirror>();

            if (r1 != null)
                result.putAll(r1);

            if (r2 != null) {
                // Need to be careful about overlap
                for (String key : r2.keySet()) {
                    if (!result.containsKey(key))
                        result.put(key, r2.get(key));
                    else if (!result.get(key).equals(r2.get(key)))
                        result.put(key, READONLY);
                }
            }
            return result;
        }

        public Map<String, AnnotationMirror> visit(Iterable<? extends AnnotatedTypeMirror> types,
                Iterable<? extends AnnotatedTypeMirror> actualTypes) {
            Map<String, AnnotationMirror> result = new HashMap<String, AnnotationMirror>();

            Iterator<? extends AnnotatedTypeMirror> itert = types.iterator();
            Iterator<? extends AnnotatedTypeMirror> itera = actualTypes.iterator();

            while (itert.hasNext() && itera.hasNext()) {
                AnnotatedTypeMirror type = itert.next();
                AnnotatedTypeMirror actualType = itera.next();
                result = reduce(result, visit(type, actualType));
            }
            return result;
        }

        @Override
        public Map<String, AnnotationMirror> visitDeclared(
                AnnotatedDeclaredType type, AnnotatedTypeMirror actualType) {

            if (actualType == null) {
                TypeElement elem = (TypeElement)type.getUnderlyingType().asElement();
                actualType = getAnnotatedType(elem);
            }
            if (actualType.getKind() != type.getKind())
                return Collections.emptyMap();

            assert actualType.getKind() == type.getKind();
            type = (AnnotatedDeclaredType)atypes.asSuper(type, actualType);
            AnnotatedDeclaredType dcType = (AnnotatedDeclaredType)actualType;

            Map<String, AnnotationMirror> result =
                new HashMap<String, AnnotationMirror>();

            if (dcType.hasAnnotation(I)) {
                String immutableString =
                    annoUtils.parseStringValue(getImmutabilityAnnotation(dcType),
                            IGJChecker.IMMUTABILITY_KEY);
                AnnotationMirror immutability = getImmutabilityAnnotation(type);
                // Assertion failes some times
                assert immutability != null;
                if (!immutability.equals(ASSIGNS_FIELDS))
                    result.put(immutableString, immutability);
            }

            result = reduce(result, visit(type.getTypeArguments(), dcType.getTypeArguments()));
            return result;
        }

        @Override
        public Map<String, AnnotationMirror> visitArray(
                AnnotatedArrayType type, AnnotatedTypeMirror actualType) {
            if (actualType == null)
                return visit(type.getComponentType(), null);

            if (type.getKind() != actualType.getKind())
                return visit(type, null);
            assert type.getKind() == actualType.getKind();
            AnnotatedArrayType arType = (AnnotatedArrayType)actualType;

            Map<String, AnnotationMirror> result =
                new HashMap<String, AnnotationMirror>();

            if (arType.hasAnnotation(I)) {
                String immutableString =
                    annoUtils.parseStringValue(getImmutabilityAnnotation(arType),
                            IGJChecker.IMMUTABILITY_KEY);
                AnnotationMirror immutability = getImmutabilityAnnotation(type);
                // Assertion failes some times
                assert immutability != null;
                if (!type.hasAnnotation(ASSIGNS_FIELDS))
                    result.put(immutableString, immutability);
            }

            result = reduce(result, visit(type.getComponentType(), arType.getComponentType()));
            return result;
        }
    }

    /**
     * Resolves {@code @I} in the type of the method type base on the method
     * invocation tree parameters.  Any unresolved {@code @I}s is resolved to a
     * place holder type.
     *
     * It resolves {@code @I} annotation in the following way:
     * <ul>
     *  <li>based on the tree receiver, done automatically through implicit
     *      invocation of
     *      {@link AnnotatedTypes#asMemberOf(AnnotatedTypeMirror, Element)}</li>
     *  <li>based on the invocation passed parameters</li>
     *  <li>if any yet unresolved immutability variables get resolved to a
     *      wildcard type</li>
     * </ul>
     */
    @Override
    public AnnotatedExecutableType methodFromUse(MethodInvocationTree tree) {
        AnnotatedExecutableType type = super.methodFromUse(tree);
        List<AnnotatedTypeMirror> arguments = atypes.getAnnotatedTypes(tree.getArguments());
        List<AnnotatedTypeMirror> requiredArgs = atypes.expandVarArgs(type, tree.getArguments());
        ImmutabilityTemplateCollector collector = new ImmutabilityTemplateCollector();
        Map<String, AnnotationMirror> matchingMapping = collector.visit(arguments, requiredArgs);
        if (!matchingMapping.isEmpty())
            new ImmutabilityResolver().visit(type, matchingMapping);

        // For finding resolved types, rather than to actually resolve immutability
        Map<String, AnnotationMirror> fromReceiver = collector.visit(getReceiver(tree));
        final Map<String, AnnotationMirror> mapping =
            collector.reduce(matchingMapping, fromReceiver);
        new AnnotatedTypeScanner<Void, Void>() {
            @Override
            public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
                if (type.hasAnnotation(I)) {
                    AnnotationMirror anno =
                        type.getAnnotation(I.class.getCanonicalName());
                    if (!mapping.containsValue(anno)) {
                        type.removeAnnotation(I);
                        type.addAnnotation(PLACE_HOLDER);
                    }
                }
                return super.visitDeclared(type, p);
            }
        }.visit(type);
        return type;
    }
}
