package checkers.igj;

import javax.annotation.processing.*;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;

import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ExpressionTree;
import com.sun.source.tree.Tree;

import checkers.basetype.BaseTypeChecker;
import checkers.igj.quals.*;
import checkers.metaquals.TypeQualifiers;
import checkers.source.SuppressWarningsKey;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.util.*;

/**
 * An annotation processor that checks a program use of IGJ mutability
 * type annotations, as specified by the FSE 2007 paper, ({@code @ReadOnly},
 * {@code @Mutable}, and {@code @Immutable}).
 *
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SuppressWarningsKey("igj")
@TypeQualifiers({ AssignsFields.class, I.class, Immutable.class, Mutable.class,
    ReadOnly.class, IGJPlaceHolder.class })
public class IGJChecker extends BaseTypeChecker {

    /** Annotation factory to create annotation mirrors */
    protected AnnotationFactory annoFactory;

    /** Supported Annotations for IGJ. Used for subtyping rules **/
    protected AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I, PLACE_HOLDER, ASSIGNS_FIELDS;

    protected IGJAnnotatedTypeFactory factory;

    protected static final String IMMUTABILITY_KEY = "value";

    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        annoFactory = new AnnotationFactory(this.env);
        READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        IMMUTABLE = annoFactory.fromName(Immutable.class.getCanonicalName());
        I = annoFactory.fromName(I.class.getCanonicalName());
        ASSIGNS_FIELDS = annoFactory.fromName(AssignsFields.class.getCanonicalName());
        PLACE_HOLDER =
            annoFactory.fromName(IGJPlaceHolder.class.getCanonicalName());
        ((GraphAnnotationRelations)this.annoRelations).setWildcard(PLACE_HOLDER);
    }

    @Override
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror use) {
        if (elemType.hasAnnotation(I) || use.hasAnnotation(READONLY))
            return true;
        return super.isValidUse(elemType, use);
    }

    @Override
    protected TypeRelations createTypeRelations() {
        return new TypeRelations(annoRelations) {
            @Override
            protected boolean compareTypeArguments(AnnotatedDeclaredType lhs, AnnotatedDeclaredType rhs) {
                if (!lhs.hasAnnotation(READONLY) && !lhs.hasAnnotation(IMMUTABLE))
                    return super.compareTypeArguments(lhs, rhs);

                if (!lhs.getTypeArguments().isEmpty() && !rhs.getTypeArguments().isEmpty()) {
                    boolean result = true;
                    for (int i = 0; i < lhs.getTypeArguments().size(); ++i) {
                        result &= isSubtype(lhs.getTypeArguments().get(i), rhs.getTypeArguments().get(i));
                    }
                    return result;
                }
                return true;

            }
            @Override
            protected boolean compareTypeArg(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
                return lhs.hasAnnotation(PLACE_HOLDER) || rhs.hasAnnotation(PLACE_HOLDER) ||
                    super.compareTypeArg(lhs, rhs);
            }
        };
    }

    /**
     * Checks whether the assignment variable is a non-assignable field, and
     * returns false if so, true otherwise.
     *
     * A field is assignable if it is
     * <ul>
     *  <li>a static field</li>
     *  <li>marked {@link Assignable}</li>
     *  <li>accessed through a mutable reference</li>
     *  <li>reassigned with an {@link AssignsFields} method and owned by
     *      'this'</li>
     * </ul>
     *
     */
    @Override
    public boolean isAssignable(AnnotatedTypeMirror varType, Tree varTree) {
        if (varTree.getKind() == Tree.Kind.VARIABLE)
            return true;

        Element varElement = InternalUtils.symbol(varTree);
        if (varElement == null || !varElement.getKind().isField()
                || ElementUtils.isStatic(varElement)
                || varElement.getAnnotation(Assignable.class) != null)
            return true;

        ExpressionTree expTree = (ExpressionTree)varTree;
        AnnotatedTypeMirror receiver = factory.getReceiver(expTree);

        boolean isAssignable =
            receiver.hasAnnotation(MUTABLE) ||
            (receiver.hasAnnotation(ASSIGNS_FIELDS)
                    && TreeUtils.isSelfAccess(expTree));

        return isAssignable;
    }

    @Override
    public IGJAnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        factory = new IGJAnnotatedTypeFactory(this, root);
        return factory;
    }
}
